import os
import errno
import re
import shutil
import struct
import logging
from io import BytesIO
from typing import Dict, List
from PIL import Image

from .storage import PatchVersion
from .wad import Wad
from .binfile import BinFile
from .sknfile import SknFile
from .rstfile import RstFile, get_hashfile as get_rsthashfile, key_to_hash as key_to_rsthash
from .tools import (
    BinaryParser,
    convert_cdragon_path,
    json_dump,
    write_file_or_remove,
    write_dir_or_remove,
    json_dumps
)

logger = logging.getLogger(__name__)


def paths_to_tree(paths):
    """Reduce an iterable of paths into nested mappings

    For leafs, value if set to None.
    There must not be paths empty parts, leading or trailing slashes.

    For instance: ['a/x/1', 'a/2']
    Is reduced to: {'a': {'x': {'1': None}, '2': None}}
    """

    tree = {}
    for path in paths:
        *parents, leaf = path.split('/')
        subtree = tree
        for parent in parents:
            subtree = subtree.setdefault(parent, {})
        subtree[leaf] = None
    return tree

def reduce_common_trees(parts, tree1, tree2, excludes):
    """Recursive method for reducing paths"""
    if tree1 is None or (tree1 == tree2 and excludes is None):
        # leaf or common, non-excluded subtree
        yield '/'.join(parts)
        return
    for name in tree1:
        # non-common subtree, compare each subtree
        # tree2[name] must exist, since tree1 must be a subtree of tree2
        yield from reduce_common_trees(parts + [name], tree1[name], tree2[name], None if excludes is None else excludes.get(name))

def reduce_common_paths(paths1, paths2, excludes):
    """Compare paths lists and return the most common subpaths

    Reduce directories in paths1 that are the same in paths2 so that the
    returned list of paths are common in paths1 and paths2.
    All paths in paths1 must exist in paths2.

    If paths lists are identical, return a list of root's subdirs.
    """

    tree1 = paths_to_tree(paths1)
    tree2 = paths_to_tree(paths2)
    tree_excludes = paths_to_tree(excludes)
    ret = list(reduce_common_trees([], tree1, tree2, tree_excludes))
    if len(ret) == 1 and ret[0] == '':
        # trees are identical
        return list(tree1)
    return ret


class Exporter:
    """Export files and WADs to a directory"""

    def __init__(self, output: str):
        self.output = os.path.normpath(output)
        self.wads: Dict[str, Wad] = {}  # {export_path: Wad}
        self.plain_files: Dict[str, str] = {}  # {export_path: path}
        self.converters: List[FileConverter] = []


    def _exported_paths(self):
        """Generate paths of all exported files, excluding converted paths"""
        yield from self.plain_files
        for wad in self.wads.values():
            yield from (wf.path for wf in wad.files)

    def unknown_hashes(self):
        """Yield all unknown hashes from WAD files"""
        for wad in self.wads.values():
            yield from (wf.path_hash for wf in wad.files if not wf.path)

    def all_converted_exported_paths(self):
        """Generate paths of all exported files, including converted paths"""
        for path in self._exported_paths():
            yield from self.converted_exported_paths(path)

    def converted_exported_paths(self, path):
        """Generate all paths that the given path will be exported as"""
        converter = self._get_converter(path)
        yield from converter.converted_paths(path)

    def walk_output_dir(self, skip_recurse=None):
        """Generate a list of files on disk (even if not in exported files)

        Don't recurse into directories in `skip_recurse`.
        Generate paths with forward slashes on all platforms.
        """
        # os.walk() handles symlinked directories as directories
        # due to this, it's simpler (and faster) to recurse ourselves
        if not os.path.exists(self.output):
            return
        if skip_recurse is None:
            skip_recurse = []
        to_visit = ['']
        while to_visit:
            base = to_visit.pop()
            with os.scandir(f"{self.output}/{base}") as scan_it:
                for entry in scan_it:
                    if entry.is_symlink() or entry.is_file(follow_symlinks=False):
                        yield f"{base}{entry.name}"
                    elif entry.is_dir():
                        path = f"{base}{entry.name}"
                        if path not in skip_recurse:
                            to_visit.append(f"{path}/")


    def add_path(self, source_path, export_path):
        """Add a path to export

        source_path is the full path to the file to export.
        export_path is the relative export path, it serves as key for
        filtering, etc.
        """

        if source_path.endswith('.wad') or source_path.endswith('.wad.client'):
            wad = Wad(source_path)
            # remove file redirections
            wad.files = [wf for wf in wad.files if wf.type != 2]
            wad.guess_extensions()
            wad.sanitize_paths()
            self.wads[export_path] = wad
        else:
            self.plain_files[export_path] = source_path

    def add_patch_files(self, patch):
        """Add files to export from a patch"""

        logger.info(f"add list of files to extract for patch {patch.version}")

        for elem in patch.latest().elements:
            elem.download(langs=True)

        # add files to export
        for elem in patch.latest().elements:
            for src, dst in elem.paths(langs=True):
                self.add_path(src, dst)

    def filter_path(self, source_path, export_path):
        """Remove files that are in the provided path

        Paths have the same meaning as for add_path().

        WAD files are first compared by source path, then by file's sha256.
        Plain files are simply removed.
        """

        if export_path in self.plain_files:
            del self.plain_files[export_path]
        elif source_path.endswith('.wad') or source_path.endswith('.wad.client'):
            self_wad = self.wads.get(export_path)
            if self_wad is None:
                return  # not exported
            if self_wad.path == source_path:
                # same path: WADs are identical
                logger.debug(f"filter identical WAD file: {source_path}")
                del self.wads[export_path]
            else:
                # compare the sha256 hashes to find the common files
                # don't resolve hashes: we just need the sha256
                logger.debug(f"filter modified WAD file: {source_path}")
                other_wad = Wad(source_path, hashes={})
                other_sha256 = {wf.path_hash: wf.sha256 for wf in other_wad.files}
                # change the files from the wad so it only extract these
                self_wad.files = [wf for wf in self_wad.files if wf.sha256 != other_sha256.get(wf.path_hash)]
                if not self_wad.files:
                    del self.wads[export_path]

    def filter_export_paths(self, predicate):
        """Filter paths to export using a predicate

        Unknown files are filtered out if None is filtered out.
        """

        self.plain_files = {k: v for k, v in self.plain_files.items() if predicate(k)}
        emptied = []
        for path, wad in self.wads.items():
            wad.files = [wf for wf in wad.files if predicate(wf.path)]
            if not wad.files:
                emptied.append(path)
        for path in emptied:
            del self.wads[path]

    def filter_exporter(self, other):
        """Remove files that are exported by another exporter"""

        for path, other_storage_path in other.plain_files.items():
            if self.plain_files.get(path) == other_storage_path:
                # same storage path: files are identical
                logger.debug(f"filter identical plain file: {path}")
                del self.plain_files[path]

        for path, other_wad in other.wads.items():
            self_wad = self.wads.get(path)
            if self_wad is None:
                continue  # not exported
            if self_wad.path == other_wad.path:
                # same path: WADs are identical
                logger.debug(f"filter identical WAD file: {path}")
                del self.wads[path]
            else:
                # compare the sha256 hashes to find the common files
                # don't resolve hashes: we just need the sha256
                logger.debug(f"filter modified WAD file: {path}")
                other_sha256 = {wf.path_hash: wf.sha256 for wf in other_wad.files}
                # change the files from the wad so it only extract these
                self_wad.files = [wf for wf in self_wad.files if wf.sha256 != other_sha256.get(wf.path_hash)]
                if not self_wad.files:
                    del self.wads[path]

    def export(self, overwrite=True):
        """Export files to the output

        If overwrite is False, don't extract files that already exist on disk.
        """

        logger.info(f"export plain files ({len(self.plain_files)})")
        for export_path, source_path in self.plain_files.items():
            self._export_plain_file(export_path, source_path, overwrite)

        for wad in self.wads.values():
            self._export_wad(wad, overwrite)

    def clean_output_dir(self, kept_files, kept_symlinks):
        """Remove regular files (or directories) and symlinks from output, except given ones

        This method is intended to be used to clean-up files that should not be
        extracted/symlinked. Parent directories are removed (if empty).
        Note: symlinks are assumed to point to the right location.
        """

        # collect files to remove
        # note: empty directories are not removed
        trees_to_remove = []
        files_to_remove = []
        for path in self.walk_output_dir(kept_files):
            full_path = os.path.join(self.output, path)
            if os.path.islink(full_path):
                if path not in kept_symlinks:
                    files_to_remove.append(full_path)
            else:
                if path not in kept_files:
                    if os.path.isdir(full_path):
                        trees_to_remove.append(full_path)
                    else:
                        files_to_remove.append(full_path)

        dirs_to_remove = set()
        for path in files_to_remove:
            logger.info(f"remove extra file or symlink: {path}")
            os.remove(path)
            dirs_to_remove.add(os.path.dirname(path))
        for path in trees_to_remove:
            logger.info(f"remove extra directory: {path}")
            shutil.rmtree(path)
            dirs_to_remove.add(os.path.dirname(path))

        for path in dirs_to_remove:
            try:
                os.removedirs(path)
            except OSError:
                pass


    def _get_converter(self, path):
        """Get converter that handles the given path, or None"""
        for converter in self.converters:
            if converter.is_handled(path):
                return converter
        return CopyConverter.singleton

    def _export_plain_file(self, export_path, source_path, overwrite=True):
        """Export a plain file"""

        converter = self._get_converter(export_path)
        if not overwrite and converter.converted_paths_exist(self.output, export_path):
            return

        try:
            with open(source_path, 'rb') as fin:
                converter.convert(fin, self.output, export_path)
        except FileConversionError as e:
            logger.warning(f"cannot convert file '{source_path}': {e}")

    def _export_wad(self, wad, overwrite=True):
        logger.info(f"export {wad.path} ({len(wad.files)})")
        # similar to Wad.extract()
        # unknown files are skipped
        with open(wad.path, 'rb') as fwad:
            for wadfile in wad.files:
                if wadfile.path is None:
                    continue

                converter = self._get_converter(wadfile.path)
                if not overwrite and converter.converted_paths_exist(self.output, wadfile.path):
                    continue

                data = wad.read_file_data(fwad, wadfile)
                if data is None:
                    continue

                try:
                    converter.convert(BytesIO(data), self.output, wadfile.path)
                except FileConversionError as e:
                    logger.warning(f"cannot convert file '{wadfile.path}': {e}")
                except OSError as e:
                    # Path components longer than 255 are not supported, ignore such files
                    if e.errno in (errno.EINVAL, errno.ENAMETOOLONG):
                        logger.warning(f"ignore file with invalid path: {wadfile.path}")
                    else:
                        raise


class CdragonRawPatchExporter:
    """Export a single patch, as on raw.communitydragon.org

    Handle symlinking of previous patch files, write list of unknown hashes,
    convert files, etc.
    """

    def __init__(self, output, patch, prev_patch=None, symlinks=None):
        self.output = os.path.normpath(output)
        self.patch = patch
        self.prev_patch = prev_patch
        if symlinks is None:
            self.create_symlinks = prev_patch is not None
        else:
            if symlinks and not prev_patch:
                raise ValueError("cannot create symlinks without a previous patch")
            self.create_symlinks = symlinks

    def process(self, overwrite=True):
        exporter = self._create_exporter(self.patch)

        # collect unknown hashes before resolving and filtering anything
        unknown_hashes = sorted(exporter.unknown_hashes())

        logger.info(f"filter and transform exported files for patch {self.patch.version}")
        self._transform_exported_files(exporter, self.patch.version)

        if self.prev_patch:
            prev_exporter = self._create_exporter(self.prev_patch)
            self._transform_exported_files(prev_exporter, self.prev_patch.version)

        # collect list of paths to extract (new ones, previous ones)
        new_paths = set(exporter.all_converted_exported_paths())
        symlinked_paths = None
        changed_paths = new_paths  # default: use all new paths
        if self.prev_patch:
            prev_paths = set(prev_exporter.all_converted_exported_paths())
            # filter out files from previous patch
            exporter.filter_exporter(prev_exporter)
            # collect a list of new paths to actually extract (changed ones)
            if self.create_symlinks:
                changed_paths = set(exporter.all_converted_exported_paths())
                # build a list of symlinks
                # note: Game files contain some duplicates which will appear in several WAD files.
                # A new version of any duplicate will override any unmodified one.
                symlinked_paths = reduce_common_paths(new_paths - changed_paths, prev_paths, changed_paths)

        # exclude generated "cdragon" files
        changed_paths.add("cdragon")
        exporter.clean_output_dir(changed_paths, set(symlinked_paths or []))

        # extract files, create symlinks if needed
        exporter.export(overwrite=overwrite)
        if symlinked_paths:
            self._create_symlinks(symlinked_paths)

        # write additional txt files
        os.makedirs(os.path.join(self.output, "cdragon"), exist_ok=True)

        if symlinked_paths:
            with open(os.path.join(self.output, "cdragon/files.links.txt"), 'w', newline='\n') as f:
                for link in sorted(symlinked_paths):
                    print(link, file=f)

        with open(os.path.join(self.output, "cdragon/files.unknown.txt"), 'w', newline='\n') as f:
            for h in unknown_hashes:
                print(f"{h:016x}", file=f)

        with open(os.path.join(self.output, "cdragon/files.exported.txt"), 'w', newline='\n') as f:
            for path in sorted(new_paths):
                print(path, file=f)

        logger.info("export TFT data files")
        self.export_tft_data()
        logger.info("export Arena data files")
        self.export_arena_data()

    def export_tft_data(self):
        if self.patch.version != 'main' and self.patch.version < PatchVersion('9.14'):
            return  # no supported TFT data before 9.14
        # don't import in module to be able to execute tftdata module
        game_version = self.patch.version.as_int()
        from .tftdata import TftTransformer
        transformer = TftTransformer(os.path.join(self.output, "game"), game_version)
        transformer.export(os.path.join(self.output, "cdragon/tft"), langs=None)

    def export_arena_data(self):
        if self.patch.version != 'main' and self.patch.version < PatchVersion('13.14'):
            return  # no supported Arena data before 13.14
        # don't import in module to be able to execute arenadata module
        from .arenadata import ArenaTransformer
        transformer = ArenaTransformer(os.path.join(self.output, "game"))
        transformer.export(os.path.join(self.output, "cdragon/arena"), langs=None)

    def _create_exporter(self, patch):
        game_version = patch.version.as_int()
        exporter = Exporter(self.output)
        exporter.converters = [
            ImageConverter(('.dds', '.tga')),
            TexConverter(),
            AtlasInfoConverter(re.compile(r'game/clientstates/.*\.cdtb$|game/assets/items/icons2d/autoatlas/.*/atlas_info\.bin$')),
            BinConverter(re.compile(r'game/.*\.bin$'), game_version),
            SknConverter(),
            RstConverter(re.compile(r'game/(?:.*/)?data/menu/.*\.(txt|stringtable)$'), game_version),
        ]
        exporter.add_patch_files(patch)
        return exporter

    @staticmethod
    def _transform_exported_files(exporter: Exporter, patch_version):
        """Filter exported files, resolve unknowns, etc."""

        # resolve unknowns paths
        for path, wad in exporter.wads.items():
            unknown_path = "unknown"
            # league_client: extract unknown files under plugin directory
            m = re.search(r'^(plugins/rcp-.+?)/[^/]*assets\.wad$', path, re.I)
            if m is not None:
                unknown_path = f"{m.group(1).lower()}/unknown"
            wad.set_unknown_paths(unknown_path)

        # don't export executables
        exporter.filter_export_paths(lambda p: p and not p.endswith('.exe') and not p.endswith('.dll'))

        # Remove 'description.json' files from plain files.
        # They may also also be in WADs (slightly different though), which may
        # result into files being both extracted and symlinked.
        # Just use WAD ones, even if it leads to not having a description.json
        # at all. These files are not needed anyway.
        exporter.plain_files = {k: v for k, v in exporter.plain_files.items() if not k.endswith('/description.json')}

        # game WADs:
        # - keep images and skin files
        # - keep .bin files, except 'data/*_skins_*.bin' files
        # - keep .txt files (some contain useful data)
        # - keep font files
        # - add 'game/' prefix to export path
        def filter_path(exporter, path):
            keep = False
            for exported_path in exporter.converted_exported_paths(path):
                _, ext = os.path.splitext(exported_path)
                if ext == '.bin':
                    return '_skins_' not in exported_path
                else:
                    keep |= ext in ('.png', '.skn', '.txt', '.stringtable', '.json', '.ttf', '.otf')
            return keep

        for path, wad in exporter.wads.items():
            if path.endswith('.wad.client'):
                subdir = "game"
                # Export language-specific files to a dedicated subdir to avoid conflicts
                # Don't do it for patches prior 14.4, to avoid breaking URLs, and possibly our own code
                if patch_version == "main" or patch_version >= PatchVersion("14.4"):
                    if lang_match := re.search(r"\.(.._..)\.wad\.client$", path):
                        subdir = f"game/{lang_match.group(1).lower()}"
                for wf in wad.files:
                    wf.path = f"{subdir}/{wf.path}"
                wad.files = [wf for wf in wad.files if filter_path(exporter, wf.path)]
        # remove emptied WADs
        for path, wad in list(exporter.wads.items()):
            if not wad.files:
                del exporter.wads[path]


    def _create_symlinks(self, symlinks):
        if not symlinks:
            return
        dst_output = self.output
        src_output = os.path.join(os.path.dirname(self.output), str(self.prev_patch.version))

        logger.info(f"creating symlinks for patch {self.patch.version}")
        for link in symlinks:
            dst = os.path.join(dst_output, link)
            if os.path.lexists(dst):
                if not os.path.islink(dst):
                    raise RuntimeError(f"symlink target already exists: {dst}")
                continue  # already set
            dst_dir = os.path.dirname(dst)
            os.makedirs(dst_dir, exist_ok=True)
            src = os.path.relpath(os.path.realpath(os.path.join(src_output, link)), os.path.realpath(dst_dir))
            logger.info(f"create symlink {dst}")
            try:
                os.symlink(src, dst)
            except OSError as e:
                # Path components longer than 255 are not supported, ignore such files
                if e.errno in (errno.EINVAL, errno.ENAMETOOLONG):
                    logger.warning(f"ignore symlink with invalid path: {dst}")
                else:
                    raise

    @classmethod
    def from_directory(cls, storage, output: str, first: PatchVersion=None, symlinks=None):
        """Handle export of multiple patchs in the same directory

        Exporter for the most oldest patch is returned first.
        """

        #XXX If directories of intermediate patches don't exist, they will
        # be fully extracted. This should be checked.

        # collect all version subdirectories
        versions = set()
        for path in os.listdir(output):
            if not os.path.isdir(os.path.join(output, path)):
                continue
            try:
                version = PatchVersion(path)
            except (ValueError, TypeError):
                continue
            versions.add(version)

        if not versions:
            return []  # no version directory found

        # get patches from versions (latest to oldest)
        patches = []
        for patch in storage.patches(stored=True):
            if first and patch.version < first:
                break
            if patch.version in versions:
                patches.append(patch)
                versions.remove(patch.version)
                if not versions:
                    break
        else:
            raise ValueError(f"versions not found: {versions!r}")

        # create patch exporters (latest to oldest)
        exporters = []
        for patch, previous_patch in zip(patches, patches[1:] + [None]):
            patch_output = os.path.join(output, str(patch.version))
            exporters.append(cls(patch_output, patch, previous_patch, symlinks=symlinks if previous_patch else False))
        return exporters[::-1]


class FileConverter:
    """Base class for file conversions

    Each single file can be converted to one or multiple files and/or
    directories (as yielded by `converted_paths()`).
    """

    def is_handled(self, path):
        """Return whether the path is handled"""
        raise NotImplementedError()

    def converted_paths(self, path):
        """Generate paths converted from `path`

        `self.handled(path) is True` can be assumed.
        """
        raise NotImplementedError()

    def converted_paths_exist(self, output, path):
        """Return True if all converted paths exist"""
        return all(os.path.lexists(os.path.join(output, p)) for p in self.converted_paths(path))

    def convert(self, fin, output, path):
        """Convert a source file object

        `output` is the root output directory.
        `path` is the relative path of the converted file.

        Implementations should use `write_file_or_remove()` and
        `write_dir_or_remove()` to ensure files are properly removed on error.

        A `FileConversionError` should be raised on conversion error.

        `self.handled(path) is True` can be assumed.
        """
        raise NotImplementedError()


class FileConversionError(RuntimeError):
    pass

class CopyConverter(FileConverter):
    """Converter that copy as-is (no actual conversion)"""

    def is_handled(self, path):
        return True

    def converted_paths(self, path):
        yield path

    def convert(self, fin, output, path):
        output_path = os.path.join(output, path)
        with write_file_or_remove(output_path) as fout:
            shutil.copyfileobj(fin, fout)

# use as a singleton, to avoid multiple instanciations for nothing
CopyConverter.singleton = CopyConverter()

class ImageConverter(FileConverter):
    def __init__(self, extensions):
        self.extensions = extensions

    def is_handled(self, path):
        return os.path.splitext(path)[1] in self.extensions

    def converted_paths(self, path):
        yield os.path.splitext(path)[0] + '.png'

    def convert(self, fin, output, path):
        output_path = os.path.join(output, os.path.splitext(path)[0] + '.png')
        with write_file_or_remove(output_path) as fout:
            try:
                im = Image.open(fin)
                im.save(fout)
            except (OSError, NotImplementedError):
                # "OSError: cannot identify image file" happen for some files with a wrong extension
                raise FileConversionError("cannot convert image to PNG")

class TexConverter(FileConverter):
    def __init__(self):
        pass

    def is_handled(self, path):
        return path.endswith('.tex')

    def converted_paths(self, path):
        yield os.path.splitext(path)[0] + '.png'

    def convert(self, fin, output, path):
        output_path = os.path.join(output, os.path.splitext(path)[0] + '.png')
        fdds = BytesIO(self.tex_to_dds(fin.read()))
        with write_file_or_remove(output_path) as fout:
            try:
                im = Image.open(fdds)
                im.save(fout)
            except (OSError, NotImplementedError):
                raise FileConversionError("cannot convert image to PNG")

    @staticmethod
    def tex_to_dds(data):
        # Parse TEX header
        if len(data) < 12 or data[:4] != b'TEX\0':
            raise FileConversionError("invalid TEX file")
        _, width, height, format, has_mipmaps = struct.unpack('<4sHHxBx?', data[:12])

        if format == 0x0a:  # DXT1
            ddspf = struct.pack('<LL4s20x', 32, 0x4, b'DXT1')
        elif format == 0x0c:  # DXT5
            ddspf = struct.pack('<LL4s20x', 32, 0x4, b'DXT5')
        elif format == 0x14:  # BGRA8
            ddspf = struct.pack('<LL4x5L', 32, 0x41, 8*4, 0x00ff0000, 0x0000ff00, 0x000000ff, 0xff000000)
        else:
            raise FileConversionError(f"unsupported TEX format: {format:x}")

        if has_mipmaps:
            # Note: only convert the largest mipmap

            if format == 0x0a:  # DXT1
                block_size = 4
                bytes_per_block = 8
            elif format == 0x0c:  # DXT5
                block_size = 4
                bytes_per_block = 16
            elif format == 0x14:  # BGRA8
                block_size = 1
                bytes_per_block = 4

            # Find mipmap count
            n = max(width, height)
            mipmap_count = 0
            while n > 0:
                mipmap_count += 1
                n >>= 1

            block_width = (width + block_size - 1) // block_size
            block_height = (height + block_size - 1) // block_size
            mipmap_size = bytes_per_block * block_width * block_height
            pixels = data[-mipmap_size:]
        else:
            pixels = data[12:]

        dds_header = struct.pack('<4s4L56x32sL16x', b'DDS ', 124, 0x1007, height, width, ddspf, 0x1000)
        return dds_header + pixels

class BinConverter(FileConverter):
    def __init__(self, regex, btype_version=None):
        self.regex = regex
        self.btype_version = btype_version

    def is_handled(self, path):
        return self.regex.search(path) is not None

    def converted_paths(self, path):
        yield path
        yield path + '.json'

    def convert(self, fin, output, path):
        output_path = os.path.join(output, path)
        with write_file_or_remove(output_path) as fout:
            shutil.copyfileobj(fin, fout)
        with write_file_or_remove(output_path + '.json') as fout:
            try:
                binfile = BinFile(output_path, btype_version=self.btype_version)
            except ValueError as e:
                raise FileConversionError(f"failed to parse bin file: {e}")
            fout.write(json_dumps(binfile.to_serializable()).encode('ascii'))

class SknConverter(FileConverter):
    def __init__(self):
        pass

    def is_handled(self, path):
        return path.endswith('.skn')

    def converted_paths(self, path):
        yield path
        yield os.path.splitext(path)[0]

    def convert(self, fin, output, path):
        output_path = os.path.join(output, path)
        with write_file_or_remove(output_path) as fout:
            shutil.copyfileobj(fin, fout)
        obj_output_path = os.path.join(output, os.path.splitext(path)[0])
        shutil.rmtree(obj_output_path, ignore_errors=True)
        with write_dir_or_remove(obj_output_path):
            sknfile = SknFile(output_path)
            for entry in sknfile.entries:
                name = os.path.join(obj_output_path, entry["name"] + ".obj")
                with open(name, "w") as f:
                    f.write(sknfile.to_obj(entry))

class RstConverter(FileConverter):
    def __init__(self, regex, game_version=1415):
        self.regex = regex
        self.hashes = get_rsthashfile(game_version).load()
        self.rsthash_version = game_version

    def is_handled(self, path):
        return self.regex.search(path) is not None

    def converted_paths(self, path):
        yield path
        yield path + '.json'

    def convert(self, fin, output, path):
        output_path = os.path.join(output, path)
        with write_file_or_remove(output_path) as fout:
            shutil.copyfileobj(fin, fout)

        rstfile = RstFile(output_path, self.rsthash_version)
        hashes = {key_to_rsthash(hash, bits=rstfile.hash_bits): value for hash, value in self.hashes.items()}
        rst_json = {"entries": {}, "version": rstfile.version}
        for key, value in rstfile.entries.items():
            if key in hashes:
                key = hashes[key]
            else:
                key = f"{{{key:010x}}}"
            rst_json["entries"][key] = value

        with write_file_or_remove(output_path + '.json', False) as fout:
            json_dump(rst_json, fout, ensure_ascii=False)

class AtlasInfoConverter(FileConverter):
    def __init__(self, regex):
        self.regex = regex

    def is_handled(self, path):
        return self.regex.search(path) is not None

    def converted_paths(self, path):
        yield path + '.json'

    def convert(self, fin, output, path):
        output_path = os.path.join(output, path)

        with write_file_or_remove(output_path + ".json", False) as fout:
            json_dump(self.parse_atlasinfo(fin), fout, ensure_ascii=False)

    @staticmethod
    def parse_atlasinfo(f):
        parser = BinaryParser(f)

        atlas_count, = parser.unpack("<L")
        atlas_paths = [convert_cdragon_path(parser.unpack_string()) for _ in range(atlas_count)]
        atlas_info = {}

        texture_count, = parser.unpack("<L")
        for _ in range(texture_count):
            texture_name = parser.unpack_string()
            startX, startY, endX, endY, atlas_index = parser.unpack("<ffffL")
            atlas_info[texture_name] = {"atlasPath": atlas_paths[atlas_index], "startX": startX, "startY": startY, "endX": endX, "endY": endY}

        return atlas_info



import os
import re
import itertools
import json
import glob
from contextlib import contextmanager
from typing import List, Tuple, Union, Optional, Generator
import logging
import requests
import hachoir.parser
import hachoir.metadata

from .tools import write_file_or_remove

logger = logging.getLogger(__name__)


class BaseVersion:
    """Base wrapper class for version strings

    Version instances are comparable and hashable.
    """

    def __init__(self, v: Union[str, tuple]):
        if isinstance(v, str):
            self.s = v
            self.t = tuple(int(x) for x in v.split('.'))
        elif isinstance(v, tuple):
            self.s = '.'.join(str(x) for x in v)
            self.t = v
        else:
            raise TypeError(v)

    def __repr__(self):
        return f"{self.__class__.__qualname__}({self.s!r})"

    def __str__(self):
        return self.s

    def __lt__(self, other):
        return self.t < other.t

    def __le__(self, other):
        return self.t <= other.t

    def __gt__(self, other):
        return self.t > other.t

    def __ge__(self, other):
        return self.t >= other.t

    def __eq__(self, other):
        # allow to compare with any other version instance, string or tuple
        if isinstance(other, BaseVersion):
            return self.s == other.s
        elif isinstance(other, str):
            return self.s == other
        elif isinstance(other, tuple):
            return self.t == other
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.s)


class PatchVersion(BaseVersion):
    """Wrapper class for patch version numbers

    Patch versions are X.Y, where X is the season and Y the patch number, starting at 1.
    Mid-update patches (e.g. 8.23.2) are not handled.

    For PBE, the special version `main` is used.
    """

    def __init__(self, v: Union[str, tuple]):
        if isinstance(v, str) and v == "main":
            # make it comparable, but only with itself
            self.s = self.t = v
        else:
            super().__init__(v)
            # support more numbers but truncate to 2
            if len(self.t) > 2:
                self.t = self.t[:2]
                self.s = '.'.join(str(x) for x in self.t)
            assert len(self.t) == 2, "invalid patch version format"

    def as_int(self):
        if self.t == "main":
            return 9999 # return highest possible number
        else:
            v0, v1 = self.t
            return v0 * 100 + v1

class RequestStreamReader:
    """Wrapper for reading data from stream request"""

    DOWNLOAD_CHUNK_SIZE = 10 * 1024**2

    def __init__(self, r):
        self.it = r.iter_content(self.DOWNLOAD_CHUNK_SIZE)
        self.pos = 0
        self.buf = b''

    def copy(self, writer, n):
        """Read n bytes and write them using writer"""
        self.pos += n
        while n:
            if n <= len(self.buf):
                if writer:
                    writer(self.buf[:n])
                self.buf = self.buf[n:]
                return
            if writer and self.buf:
                writer(self.buf)
            n -= len(self.buf)
            self.buf = next(self.it)

    def skip(self, n):
        """Skip n bytes"""
        self.copy(None, n)

    def skip_to(self, pos):
        assert self.pos <= pos
        self.skip(pos - self.pos)


# registered storage classes
_storage_registry = {}

def load_storage_conf(path):
    with open(path) as f:
        conf = json.load(f)
    if 'type' not in conf:
        raise ValueError("storage configuration file must define its 'type'")
    conf['path'] = os.path.normpath(os.path.join(os.path.dirname(path), conf.get('path', '.')))
    return conf

def storage_conf_from_path(path):
    """Parse a path as supported by Storage.from_path()

    The returned conf is guaranteed to have a `type` and a `path` entries.
    """
    if os.path.isdir(path):
        conf_path = os.path.join(path, 'cdtb.storage.conf')
        if os.path.isfile(conf_path):
            return load_storage_conf(conf_path)
        else:
            conf = guess_storage_conf(path)
            if conf is None:
                raise ValueError(f"cannot guess storage configuration from '{path}'")
            return conf
    elif os.path.isfile(path):
        return load_storage_conf(path)
    elif ':' in path:
        storage_type, storage_path = path.split(':', 1)
        return {'type': storage_type, 'path': storage_path}
    else:
        return {'type': 'patcher', 'path': path}

def guess_storage_conf(path):
    """Try to guess storage configuration from path"""

    if os.path.isdir(os.path.join(path, 'channels')):
        return {'type': 'patcher', 'path': path}
    return None


class StorageRegister(type):
    """Metaclass to register storage types"""

    def __new__(mcs, name, bases, class_dict):
        cls = type.__new__(mcs, name, bases, class_dict)
        if cls.storage_type is not None:
            _storage_registry[cls.storage_type] = cls
        return cls

class Storage(metaclass=StorageRegister):
    """
    Download and store game and client files

    Each storage is basically a directory in which files are downloaded and
    extracted if needed. Each storage type can define configuration options.
    """

    storage_type = None

    def __init__(self, path, url):
        self.path = path
        self.url = url
        self.s = requests.session()

    @staticmethod
    def from_path(path):
        """Return a storage from a path

        `path` can points to:
        - a storage configuration file
        - a directory containing a `cdtb.storage.conf` file
        - a directory (storage configuration will be guessed, if possible)
        - `type:dir_path` string
        """

        conf = storage_conf_from_path(path)
        if conf is None:
            raise ValueError(f"cannot retrieve storage configuration from '{path}'")
        return Storage.from_conf(conf)

    @staticmethod
    def from_conf(conf):
        try:
            cls = _storage_registry[conf['type']]
        except KeyError:
            raise ValueError(f"unknown storage type: {conf['type']}")
        return cls.from_conf_data(conf)

    @classmethod
    def from_conf_data(cls, conf):
        raise NotImplementedError()

    def request_get(self, path, **kwargs) -> requests.Response:
        """Request a path, returns a requests.Response object"""
        return self.s.get(self.url + path, **kwargs)

    def request_text(self, path) -> str:
        """Request a path, return content as text"""
        r = self.request_get(path)
        r.raise_for_status()
        r.encoding = 'utf-8'
        return r.text

    def fspath(self, path) -> str:
        """Return full path from a storage-relative path"""
        return os.path.join(self.path, path)

    def download(self, urlpath, path, force=False) -> None:
        """Download a path to disk
        If path is None, use urlpath's value.
        """

        if path is None:
            path = urlpath
        fspath = self.fspath(path)
        if not force and os.path.isfile(fspath):
            return

        logger.debug(f"download file: {path}")
        r = self.request_get(urlpath)
        r.raise_for_status()
        with write_file_or_remove(fspath) as f:
            f.write(r.content)

    @contextmanager
    def stream(self, urlpath) -> RequestStreamReader:
        """Request a path for streaming download"""
        with self.s.get(self.url + urlpath, stream=True) as r:
            r.raise_for_status()
            yield RequestStreamReader(r)

    def patch_elements(self, stored=False) -> Generator['PatchElement', None, None]:
        """Generate patch elements, sorted from the latest one

        If stored is True, only elements already in storage are used (to avoid
        downloading new files).

        Versions are generated so the caller can stop iterating when needed
        versions have been retrieved, avoiding to fetch all solutions.

        Note: patch versions are assumed to be monotonous in successive
        solution versions (they never decrease).

        For PBE, patch version is always 'main'.
        """

        raise NotImplementedError()

    def patch_element(self, name, version=None, stored=False) -> Optional['PatchElement']:
        """Retrieve a single patch element, None if not found

        If version if None, retrieve the latest one with given name.
        """

        for e in self.patch_elements(stored=stored):
            if e.name != name:
                continue
            if version is None or e.version == version:
                return e
        return None

    def patches(self, stored=False) -> Generator['Patch', None, None]:
        """Generate patch, sorted from the latest one

        See patch_elements() for additional remarks.
        """

        for _, group in itertools.groupby(self.patch_elements(stored=stored), key=lambda e: e.version):
            # keep latest sub-patch version of each element
            elements = {}
            for elem in group:
                if elem.name not in elements:
                    elements[elem.name] = elem
            yield Patch._create(list(elements.values()))

    def patch(self, version=None, stored=False) -> Optional['Patch']:
        """Retrieve a single patch, None if not found

        If version if None, retrieve the latest one.
        """

        it = self.patches(stored=stored)
        if version is None:
            return next(it)
        for p in it:
            if p.version == version:
                return p
        return None


class PatchElement:
    """
    Element of a patch (game or client)

    This base class must not be instantiated directly.

    In methods parameters, `langs` is used to filter language-specific files
    and can have the following values:
      False -- language-independent
      True -- all languages
      lang -- single given language
      [lang, ...] -- list of given languages

    """

    # valid names
    names = ('game', 'client')

    def __init__(self, name, version: PatchVersion):
        self.name = name
        assert name in self.names
        self.version = version

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.name} {self.version}>"

    def __eq__(self, other):
        if isinstance(other, PatchElement):
            return self.name == other.name and self.version == other.version
        return False

    def __hash__(self):
        return hash((self.name, self.version))

    def download(self, langs=True):
        """Download files of this patch element to the storage"""
        raise NotImplementedError()

    def fspaths(self, langs=True) -> Generator[str, None, None]:
        """Generate the path on disk of files in the element"""
        raise NotImplementedError()

    def relpaths(self, langs=True) -> Generator[str, None, None]:
        """Generate the relative (export) path of files in the element, normalized"""
        raise NotImplementedError()

    def paths(self, langs=True) -> Generator[Tuple[str, str], None, None]:
        """Equivalent zip(fspaths(), relpaths())"""
        return zip(self.fspaths(langs=langs), self.relpaths(langs=langs))


class Patch:
    """
    A single League patch version (e.g. patch 8.1)

    This class cannot not be instantiated directly.
    Use patch() or patches() methods on Storage instances.
    """

    def __init__(self):
        raise RuntimeError("This class should not be instantiated by the user.")

    @classmethod
    def _create(cls, elements: List[PatchElement]):
        """Create a patch from its elements"""

        self = cls.__new__(cls)
        versions = {elem.version for elem in elements}
        if len(versions) != 1:
            raise ValueError("versions of patch elements mismatch")
        self.version = versions.pop()
        self.elements = elements
        return self

    def __str__(self):
        return f"patch={self.version}"

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.version}>"

    def __eq__(self, other):
        if isinstance(other, Patch):
            return self.version == other.version
        return False

    def __hash__(self):
        return hash(self.version)

    def __lt__(self, other):
        if isinstance(other, Patch):
            return self.version > other.version
        return NotImplemented

    def latest(self):
        """Return a new patch with only latest version of each element name"""
        elements = []
        for _, group in itertools.groupby(self.elements, key=lambda e: e.name):
            elements.append(max(group, key=lambda e: e.version))
        return self._create(elements)

    def download(self, langs=True):
        for elem in self.elements:
            elem.download(langs=langs)


def get_system_yaml_version(path) -> str:
    with open(path) as f:
        for line in f:
            # TODO do proper yaml parsing
            # formats: Release/X.Y or 'X.Y'
            m = re.match(r"""^ *(?:game-|)branch: .*["'/]([0-9.]+)["']?$""", line)
            if m:
                return m.group(1)
        return None


def get_exe_version(path) -> str:
    """Return version from an executable"""

    parser = hachoir.parser.createParser(path)
    metadata = hachoir.metadata.extractMetadata(parser=parser)
    return metadata.get('version')


def get_content_metadata_version(path) -> str:
    """Return branch version from content-metadata.json file"""
    with open(path) as f:
        data = json.load(f)
        m = re.match(r"^(\d+\.\d+)\.", data['version'])
        if m:
            return m.group(1)
        return None


def parse_storage_component(storage: Storage, component: str) -> Union[None, Patch, PatchElement]:
    """Parse a component string representation to patch elements"""

    m = re.match(fr'^(patch|{"|".join(PatchElement.names)})=(|[0-9]+(?:\.[0-9]+\.?)*|main)$', component)
    if not m:
        raise ValueError(f"invalid component: {component}")
    name, version = m.group(1, 2)

    if version == '':
        version, latest = None, True
    elif version.endswith('.'):
        version, latest = version.rstrip('.'), True
    else:
        latest = False

    if name == 'patch':
        patch = storage.patch(version)
        if patch is None:
            return None
        if latest:
            patch = patch.latest()
        return patch
    else:
        return storage.patch_element(name, version, stored=version is not None)



import os
import errno
import struct
import gzip
import json
import imghdr
import logging
from xxhash import xxh3_64_intdigest

from .hashes import default_hashfile
from .tools import (
    BinaryParser,
    write_file_or_remove,
    zstd_decompress,
)

def test_jpeg_photoshop(h, f):
    if h[:4] == b'\xff\xd8\xff\xe1':
        return 'jpeg'

imghdr.tests.append(test_jpeg_photoshop)


logger = logging.getLogger(__name__)


# Cache for guessed extensions.
# Caching is possible because the same hash should always have the same extension. Since guessing extension requires to
# read file data, caching will reduce I/Os
_hash_to_guessed_extensions = {}


class MalformedSubchunkError(Exception):
    """Subchunk data is invalid or doesn't match the provided subchunktoc"""

    def __init__(self, data):
        self.wad_data = data


class WadFileHeader:
    """Single file entry in a WAD archive"""

    _magic_numbers_ext = {
        b'OggS': 'ogg',
        bytes.fromhex('00010000'): 'ttf',
        bytes.fromhex('1a45dfa3'): 'webm',
        b'true': 'ttf',
        b'OTTO\0': 'otf',
        b'"use strict";': 'min.js',
        b'<template ': 'template.html',
        b'<!-- Elements -->': 'template.html',
        b'DDS ': 'dds',
        b'<svg': 'svg',
        b'PROP': 'bin',
        b'PTCH': 'bin',
        b'BKHD': 'bnk',
        b'r3d2Mesh': 'scb',
        b'r3d2anmd': 'anm',
        b'r3d2canm': 'anm',
        b'r3d2sklt': 'skl',
        b'r3d2': 'wpk',
        bytes.fromhex('33221100'): 'skn',
        b'PreLoadBuildingBlocks = {': 'preload',
        b'\x1bLuaQ\x00\x01\x04\x04': 'luabin',
        b'\x1bLuaQ\x00\x01\x04\x08': 'luabin64',
        bytes.fromhex('023d0028'): 'troybin',
        b'[ObjectBegin]': 'sco',
        b'OEGM': 'mapgeo',
        b'TEX\0': 'tex'
    }

    def __init__(self, path_hash, offset, compressed_size, size, type, duplicate=None, first_subchunk_index=None, sha256=None):
        self.path_hash = path_hash
        self.offset = offset
        self.size = size
        self.subchunk_count = (type & 0xF0) >> 4
        self.type = type & 0xF
        self.compressed_size = compressed_size
        self.duplicate = bool(duplicate)
        self.first_subchunk_index = first_subchunk_index
        self.sha256 = sha256
        # values that can be guessed
        self.path = None
        self.ext = None

    def read_data(self, f, subchunk_toc=None):
        """Retrieve (uncompressed) data from WAD file object"""

        f.seek(self.offset)
        # assume files are small enough to fit in memory
        data = f.read(self.compressed_size)
        if self.type == 0:
            return data
        elif self.type == 1:
            return gzip.decompress(data)
        elif self.type == 2:
            n, = struct.unpack('<L', data[:4])
            target = data[4:4+n].rstrip(b'\0').decode('utf-8')
            logger.debug(f"file redirection: {target}")
            return None
        elif self.type == 3:
            return zstd_decompress(data)
        elif self.type == 4:
            # Data is split into individual subchunks that may be zstd compressed
            if subchunk_toc is not None:
                chunks_data = []
                offset = 0
                for index in range(self.first_subchunk_index, self.first_subchunk_index + self.subchunk_count):
                    compressed_size, uncompressed_size, subchunk_hash = struct.unpack('<IIQ', subchunk_toc[16*index:16*(index+1)])
                    # ensure wad data matches with the subchunktoc data
                    subchunk_data = data[offset:offset+compressed_size]
                    if len(data) < offset + compressed_size or xxh3_64_intdigest(subchunk_data) != subchunk_hash:
                        raise MalformedSubchunkError(data)
                    if compressed_size == uncompressed_size:
                        # assume data is uncompressed
                        chunks_data.append(subchunk_data)
                    else:
                        chunks_data.append(zstd_decompress(subchunk_data))
                    offset += compressed_size
                return b"".join(chunks_data)
            else:
                # No subchunk TOC, try to decompress
                try:
                    return zstd_decompress(data)
                except:
                    raise MalformedSubchunkError(data)
        raise ValueError(f"unsupported file type: {self.type}")

    def extract(self, fwad, output_path, subchunk_toc=None):
        """Read data, convert it if needed, and write it to a file

        On error, partially retrieved files are removed.
        File redirections are skipped.
        """

        try:
            data = self.read_data(fwad, subchunk_toc)
        except MalformedSubchunkError:
            return
        if data is None:
            return

        try:
            with write_file_or_remove(output_path) as fout:
                fout.write(data)
        except OSError as e:
            # Path components longer than 255 are not supported, ignore such files
            # TODO: Find a better way of handling these files
            if e.errno in (errno.EINVAL, errno.ENAMETOOLONG):
                logger.warning(f"ignore file with invalid path: {self.path}")
            else:
                raise

    @staticmethod
    def guess_extension(data):
        # image type
        typ = imghdr.what(None, h=data)
        if typ == 'jpeg':
            return 'jpg'
        elif typ == 'xbm':
            pass  # some HLSL files are recognized as xbm
        elif typ is not None:
            return typ

        # json
        try:
            json.loads(data)
            return 'json'
        except (json.JSONDecodeError, UnicodeDecodeError):
            pass

        # others
        for prefix, ext in WadFileHeader._magic_numbers_ext.items():
            if data.startswith(prefix):
                return ext
        return None


class Wad:
    """A WAD archive is a file that contains other files.

    It has a header that describes the format. There are multiple
    formats that Riot uses depending on the version of the WAD file, which can be read from the header.
    The files contained in a WAD file generally are all related to one "idea".

    This class has one major purpose: to extract the individual files in a specific WAD archive for further analysis.
    """

    def __init__(self, path, hashes=None):
        self.path = path
        self.version = None
        self.files = None
        self.parse_headers()
        self.resolve_paths(hashes)
        self.load_subchunk_toc()

    def parse_headers(self):
        """Parse version and file list"""

        logger.debug(f"parse headers of {self.path}")
        with open(self.path, 'rb') as f:
            parser = BinaryParser(f)
            magic, version_major, version_minor = parser.unpack("<2sBB")
            if magic != b'RW':
                raise ValueError("invalid magic code")
            self.version = (version_major, version_minor)

            if version_major == 1:
                parser.seek(8)
            elif version_major == 2:
                parser.seek(100)
            elif version_major == 3:
                parser.seek(268)
            else:
                raise ValueError(f"unsupported WAD version: {version_major}.{version_minor}")

            entry_count, = parser.unpack("<I")

            if version_major == 1:
                self.files = [WadFileHeader(*parser.unpack("<QIIII")) for _ in range(entry_count)]
            else:
                self.files = [WadFileHeader(*parser.unpack("<QIIIB?HQ")) for _ in range(entry_count)]

    def resolve_paths(self, hashes=None):
        """Guess path of files"""

        if hashes is None:
            hashes = default_hashfile(self.path).load()
        for wadfile in self.files:
            if wadfile.path_hash in hashes:
                wadfile.path = hashes[wadfile.path_hash]
                wadfile.ext = os.path.splitext(wadfile.path)[1][1:]

    def load_subchunk_toc(self):
        """Find subchunk TOC if available and parse it"""

        for wadfile in self.files:
            if wadfile.path is None:
                continue
            if not wadfile.path.endswith(".subchunktoc"):
                continue
            with open(self.path, 'rb') as fwad:
                self.subchunk_toc = wadfile.read_data(fwad)
            break
        else:
            # Not found
            self.subchunk_toc = None

    def guess_extensions(self):
        # avoid opening the file if not needed
        unknown_ext = True
        for wadfile in self.files:
            if not wadfile.ext:
                wadfile.ext = _hash_to_guessed_extensions.get(wadfile.path_hash)
                if not wadfile.ext:
                    unknown_ext = True
        if not unknown_ext:
            return  # all extensions are known

        with open(self.path, 'rb') as f:
            for wadfile in self.files:
                if not wadfile.ext:
                    data = self.read_file_data(f, wadfile)
                    if not data:
                        continue
                    wadfile.ext = WadFileHeader.guess_extension(data)
                    _hash_to_guessed_extensions[wadfile.path_hash] = wadfile.ext

    def set_unknown_paths(self, path):
        """Set a path for files without one"""

        for wadfile in self.files:
            if not wadfile.path:
                if wadfile.ext:
                    wadfile.path = f"{path}/{wadfile.path_hash:016x}.{wadfile.ext}"
                else:
                    wadfile.path = f"{path}/{wadfile.path_hash:016x}"

    def sanitize_paths(self):
        """Sanitize paths for extract purposes; for example truncating files whose basename has a length of at least 250"""

        for wadfile in self.files:
            if wadfile.path:
                ext = os.path.splitext(wadfile.path)[1]
                if not ext:
                    # some extensionless files conflict with folder names
                    # append a custom suffix to resolve this conflict
                    ext = ".cdtb"

                    if wadfile.ext:
                        # extension was guessed, but the resolved path has no extension
                        # in this case, append the guessed extension
                        ext += f".{wadfile.ext}"

                    wadfile.path += ext

                path, filename = os.path.split(wadfile.path)
                if len(filename) >= 250:
                    wadfile.path = os.path.join(path, f"{filename[:250-17-len(ext)]}.{wadfile.path_hash:016x}{ext}")

    def extract(self, output, overwrite=True):
        """Extract WAD file

        If overwrite is False, don't extract files that already exist on disk.
        """

        logger.info(f"extracting {self.path} to {output}")

        self.sanitize_paths()
        self.set_unknown_paths("unknown")

        with open(self.path, 'rb') as fwad:
            for wadfile in self.files:
                output_path = os.path.join(output, wadfile.path)

                if not overwrite and os.path.exists(output_path):
                    logger.debug(f"skipping {wadfile.path_hash:016x} {wadfile.path} (already extracted)")
                    continue
                logger.debug(f"extracting {wadfile.path_hash:016x} {wadfile.path}")
                wadfile.extract(fwad, output_path, self.subchunk_toc)

    def read_file_data(self, fwad, wadfile):
        """Retrieve (uncompressed) data from WAD file object

        Similar to `WadFileHeader.read_data()` but use wad's subchuk information if available.
        Subchunk errors are ignored and None is returned if one happens.
        """

        try:
            return wadfile.read_data(fwad, self.subchunk_toc)
        except MalformedSubchunkError:
            return None



import os
import copy
from .storage import PatchVersion
from .binfile import BinFile
from .rstfile import RstFile
from .tools import convert_cdragon_path, json_dump, stringtable_paths


class ArenaTransformer:
    def __init__(self, input_dir, game_version=1415):
        self.input_dir = input_dir
        self.rsthash_version = game_version

    def build_template(self):
        """Parse bin data into template data"""
        map30_file = os.path.join(self.input_dir, "data", "maps", "shipping", "map30", "map30.bin")

        map30 = BinFile(map30_file)

        augments = self.parse_augments(map30)

        return {
            "augments": augments,
        }

    def export(self, output, langs=None):
        """Export Arena data for given languages

        By default (`langs` is `None`), export all available languages.
        Otherwise, export for given `xx_yy` language codes.
        """

        stringtables = stringtable_paths(self.input_dir, "lol")
        if langs is None:
            langs = list(stringtables)

        os.makedirs(output, exist_ok=True)

        template = self.build_template()
        for lang in langs:
            instance = copy.deepcopy(template)
            replacements = RstFile(stringtables[lang], self.rsthash_version)

            def replace_in_data(entry):
                for key in ("name", "desc", "tooltip"):
                    if key in entry and entry[key] in replacements:
                        entry[key] = replacements[entry[key]]

            for augment in instance["augments"]:
                replace_in_data(augment)

            with open(os.path.join(output, f"{lang}.json"), "w", encoding="utf-8") as f:
                json_dump(instance, f, indent=4, sort_keys=True, ensure_ascii=False)

    def parse_augments(self, map30):
        """Returns a list of augments"""
        augment_entries = [x for x in map30.entries if x.type == 0x6DFAB860]
        spellobject_entries = {x.path: x for x in map30.entries if x.type == "SpellObject"}

        augments = []
        for augment in augment_entries:
            if not augment.getv("enabled", True):
                continue
            augment_datavalues = {}
            augment_calculations = {}
            augment_spellobject = augment.getv(0x1418F849)
            if augment_spellobject:
                augment_spell = spellobject_entries[augment_spellobject].getv('mSpell')
                for datavalue in augment_spell.getv('mDataValues', []):
                    augment_datavalues[datavalue.getv("mName")] = datavalue.getv("mValues", [0])[0]

                #Giving raw calculations data due to not having a well defined standard
                if augment_spell.get('mSpellCalculations'):
                    augment_calculations = augment_spell.get('mSpellCalculations').to_serializable()[1]

            augments.append({
                "id": augment.getv(0x827DC19E),
                "apiName": augment.getv(0x19AE3E16),
                "name": augment.getv(0x2127EB37),
                "desc": augment.getv("DescriptionTra"),
                "tooltip": augment.getv(0x366935FC),
                "iconSmall": convert_cdragon_path(augment.getv(0x45481FB5)),
                "iconLarge": convert_cdragon_path(augment.getv(0xF1F7E50D)),
                "rarity": augment.getv("rarity", 0),
                "dataValues": augment_datavalues,
                "calculations": augment_calculations,
            })

        return augments

def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help="directory with extracted bin files")
    parser.add_argument("-o", "--output", default="arena", help="output directory")
    parser.add_argument('-V', '--patch-version', default=None,
                           help="patch version the input files belong to in the format XX.YY (default: latest patch)")
    args = parser.parse_args()

    parsed_version = PatchVersion(args.patch_version if args.patch_version else "main").as_int()

    arena_transformer = ArenaTransformer(args.input, game_version=parsed_version)
    arena_transformer.export(args.output, langs=None)

if __name__ == "__main__":
    main()
